
// TexCube.java
// Andrew Davison, April 2007, ad@fivedots.coe.psu.ac.th

/* A textured cube, centered on the origin, with sides of 
   2 units. The cube is defines as 6 square faces, with each 
   face defined by a triangle strip of two triangles
   (4 coordinates).

   The cube can be positioned and scaled, depending on its
   constructor arguments.

*/

import java.nio.*;
import javax.microedition.khronos.egl.*;
import javax.microedition.khronos.opengles.*;
import javax.microedition.lcdui.*;


public class TexCube
{
  // arrays for the cube data 
  private byte verts[] = {  // organized into 6 faces
    -1, -1, 1,   1, -1, 1,   -1, 1, 1,    1, 1, 1,   // front face
     1, -1, -1, -1, -1, -1,   1, 1, -1,  -1, 1, -1,  // back
     1, -1, 1,   1, -1, -1,   1, 1, 1,    1, 1, -1,  // right
    -1, -1, -1, -1, -1, 1,   -1, 1,-1,   -1, 1, 1,   // left
    -1, 1, 1,    1, 1, 1,    -1, 1, -1,   1, 1, -1,  // top
    -1, -1, -1,  1, -1, -1,   -1, -1, 1,  1, -1, 1   // bottom
  };

  private byte texCoords[] = {  // 4 tex coords for each face
    0,0, 1,0, 0,1, 1,1,   0,0, 1,0, 0,1, 1,1,   0,0, 1,0, 0,1, 1,1, 
    0,0, 1,0, 0,1, 1,1,   0,0, 1,0, 0,1, 1,1,   0,0, 1,0, 0,1, 1,1
  };

  private float normals[] =    // a normal for each face
    { 0, 0, 1.0f,   0, 0, -1.0f,   1.0f, 0, 0,   
     -1.0f, 0, 0,   0, 1.0f, 0,    0, -1.0f, 0  };

  private byte strips[] = { 4, 4, 4, 4, 4, 4 };  
      // a strip length for each face


  private GL10 gl;

  // buffers for the cube data
  private ByteBuffer vertsBuf;  // vertices
  private ByteBuffer tcsBuf;    // tex coords
  private ByteBuffer texBuf;    // RGB texture image

  private int imWidth, imHeight;   // dimensions of texture image
  private int texNames[];   // for the texture name

  private float xPos, yPos, zPos;  // position of cube's center
  private float scale;


  public TexCube(GL10 gl, String texFnm)
  {  this(gl, texFnm, 0, 0, 0, 1);  }


  public TexCube(GL10 gl, String texFnm, 
                        float x, float y, float z, float sc)
  // (x,y,z) is the cube's position, and sc the scaling factor
  {
    this.gl = gl;
    xPos = x; yPos = y; zPos = z;
    scale = sc;

    createCube();
    loadTexture(texFnm);

    // generate a texture name
    texNames = new int[1];
    gl.glGenTextures(1, texNames, 0); 
  }  // end of TexCube()


  private void createCube()
  // create the vertices and tex coords buffers for the cube
  {
    // create vertices buffer for cube
    vertsBuf = ByteBuffer.allocateDirect(verts.length);
    vertsBuf.put(verts).rewind();

    // 2D tex coords buffer for cube
    tcsBuf = ByteBuffer.allocateDirect(texCoords.length);
    tcsBuf.put(texCoords).rewind();
  } // end of createCube()


  private void loadTexture(String fnm)
  // load the image stored in fnm into a byte buffer
  {
    try {
      Image image = Image.createImage(fnm);
      imWidth = image.getWidth();
      imHeight = image.getHeight();
      int pixels[] = new int[imWidth];   // for a row of pixels

      texBuf = ByteBuffer.allocateDirect(imWidth * imHeight * 3);

      /* Read the image in reverse row order to flip the image
         vertically in the buffer. This will order the texture 
         correctly for OpenGL's coordinate system, where (0,0) 
         is the lower left. */
      for (int i = imHeight-1; i >= 0; i--) {
        image.getRGB(pixels, 0, imWidth, 0, i, imWidth, 1);  // get a pixel row
        for (int j = 0; j < imWidth; j++) {
          // store each pixel in the row as three bytes
          texBuf.put((byte) (pixels[j] >> 16 & 0xff));  // red
          texBuf.put((byte) (pixels[j] >> 8 & 0xff));   // green
          texBuf.put((byte) (pixels[j] >> 0 & 0xff));   // blue
        }
      }
      texBuf.rewind();
    }
    catch (Exception e) 
    {  System.out.println("Error creating texture for " + fnm); }
  }  // end of loadTexture()



  public void draw()
  // called from RotBoxGameCanvas to draw the cube
  {
    // enable the use of vertex and tex coord arrays when rendering
    gl.glEnableClientState(GL10.GL_VERTEX_ARRAY);
    gl.glEnableClientState(GL10.GL_TEXTURE_COORD_ARRAY);

    gl.glVertexPointer(3, GL10.GL_BYTE, 0, vertsBuf);  // use cube verts
    gl.glTexCoordPointer(2, GL10.GL_BYTE, 0, tcsBuf);  // use tex coords

    gl.glEnable(GL10.GL_TEXTURE_2D);   // use texturing
    setTexture();

    gl.glPushMatrix();

      gl.glTranslatef(xPos, yPos, zPos);   // move to the specified (x,y,z) pos
      if (scale != 1.0f)
        gl.glScalef(scale, scale, scale);  // uniform scaling

      int pos = 0;
      for (int i = 0; i < strips.length; i++) {  // draw each cube face
        gl.glNormal3f( normals[i*3 + 0], 
                       normals[i*3 + 1],
                       normals[i*3 + 2] );   // normal for the face
        gl.glDrawArrays(GL10.GL_TRIANGLE_STRIP, pos, strips[i]);  
                  // vertices and tex coords for the face
        pos += strips[i];
      }

    gl.glPopMatrix();

    gl.glDisableClientState(GL10.GL_VERTEX_ARRAY);
    gl.glDisableClientState(GL10.GL_TEXTURE_COORD_ARRAY); 

    gl.glDisable(GL10.GL_TEXTURE_2D);
  }  // end of drawCube()


  private void setTexture()
  {
    gl.glBindTexture(GL10.GL_TEXTURE_2D, texNames[0]);  // use the tex name

    // specify the texture for the currently bound tex name
    gl.glTexImage2D(GL10.GL_TEXTURE_2D, 0, GL10.GL_RGB, imWidth, imHeight, 0, 
                    GL10.GL_RGB, GL10.GL_UNSIGNED_BYTE, texBuf);

    // set the minification/magnification techniques
    gl.glTexParameterx(GL10.GL_TEXTURE_2D, GL10.GL_TEXTURE_MIN_FILTER, GL10.GL_LINEAR);
    gl.glTexParameterx(GL10.GL_TEXTURE_2D, GL10.GL_TEXTURE_MAG_FILTER, GL10.GL_LINEAR);
                                                                  // GL10.GL_NEAREST
  } // end of setTexture()


}  // end of TexCube class
